@@ -2,7 +2,7 @@
 
 #
 #    MiMiCPy: Python Based Tools for MiMiC
-#    Copyright (C) 2020-2021 Bharath Raghavan,
+#    Copyright (C) 2020-2023 Bharath Raghavan,
 #                            Florian Schackert
 #
 #    This file is part of MiMiCPy.
@@ -91,13 +91,13 @@ def prepqm(args):
     def selection_help():
         print("\nvalid subcommands:\n\n"
               "    o add       <selection>    add selected atoms to QM region\n"
-              "    o add-link  <selection>    add selected atoms to QM region as link atoms\n"
+              "    o add-bound  <selection>    add selected atoms to QM region as boundary atoms\n"
               "    o delete    <selection>    delete selected atoms from QM region\n"
               "    o clear                    clear all atoms from QM region\n"
               "    o view      <file-name>    print current QM region to console or file\n"
               "    o quit                     create CPMD input and GROMACS index files and quit\n"
               "    o help                     show this help message\n\n"
-              "For more information on the selection langauge please refer to the docs.\n")
+              "For more information on the selection language please refer to the docs.\n")
 
     def view(file_name=None):
         if prep.qm_atoms.empty:
@@ -113,11 +113,7 @@ def prepqm(args):
             else:
                 print(qmatoms_str)
 
-    if args.fix:
-        top = fixtop(args)
-        mpt = mimicpy.Mpt(top.molecules, top.topol_dict)
-    else:
-        mpt = get_nsa_mpt(args)
+    mpt = get_nsa_mpt(args)
 
     print('')
     loader = Loader('**Reading coordinates**')
@@ -143,7 +139,7 @@ def prepqm(args):
     readline.parse_and_bind('set editing-mode vi') # handle command history
 
     dispatch = {'add':prep.add,
-                'add-link': lambda selection: prep.add(selection, True),
+                'add-bound': lambda selection: prep.add(selection, True),
                 'delete':prep.delete,
                 'clear': prep.clear,
                 'view':view,
@@ -169,8 +165,8 @@ def prepqm(args):
     def do_prep():
         try:
             if not prep.qm_atoms.empty:
-                if args.link:
-                    prep.find_link_atoms()
+                if args.bound:
+                    prep.find_bound_atoms()
 
                 prep.get_mimic_input(args.inp, args.ndx, args.out,
                                      args.pad, args.abs, args.qma,
@@ -178,14 +174,20 @@ def prepqm(args):
 
                 if args.mdp:
                     print("Checking that {} is consistent with a MiMiC run".format(args.mdp))
-                    prepmm(argparse.Namespace(mdp=args.mdp, qma=args.qma, out=None))
+                    new_mdp = "fixed_{}".format(args.mdp)
+                    prepmm(mdp=args.mdp, qma=args.qma, out=new_mdp)
+
+                    if isfile(new_mdp):
+                        mdp = new_mdp
+                    else:
+                        mdp = args.mdp
 
                     print("\nGenerating GROMACS TPR file using {} grompp\n".format(args.gmx))
                     import subprocess
-                    subprocess.Popen("{} grompp -f {} -c {} -p {} -n {} -o {} -quiet".format(args.gmx, args.mdp,
-                                                                                        args.coords, args.top,
-                                                                                        args.ndx, args.tpr),
-                                                                                        stdout=subprocess.PIPE, shell=True).stdout.read()
+                    subprocess.Popen("{} grompp -f {} -c {} -p {} -n {} -o {} -quiet".format(args.gmx, mdp,
+                                                                                                            args.coords, args.top,
+                                                                                                            args.ndx, args.tpr),
+                                                                                                            stdout=subprocess.PIPE, shell=True).stdout.read()
                 else:
                     print("MDP file not passed! Skipping generation of GROMACS TPR file")
         except (mimicpy.utils.errors.SelectionError, mimicpy.utils.errors.MiMiCPyError) as error:
@@ -223,6 +225,16 @@ def prepqm(args):
         if check_user_input(user_input) == False:
             continue
 
+def prepmm(mdp, qma, out):
+    try:
+        mimicpy.Preparation.get_gmx_input(mdp, qma, out)
+    except FileNotFoundError as e:
+        print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
+        sys.exit(1)
+    except mimicpy.utils.errors.ScriptError as e:
+        print(e)
+        sys.exit(1)
+        
 def get_nsa_mpt(args, only_nsa=False):
     nsa_dct = {}
     if args.nsa:
@@ -256,7 +268,7 @@ def get_nsa_mpt(args, only_nsa=False):
 
     print("\n**Reading topology**\n")
     try:
-        return mimicpy.Mpt.from_file(args.top, mode='w', nonstandard_atomtypes=nsa_dct,
+        return mimicpy.Mpt.from_file(args.top, mode='r', nonstandard_atomtypes=nsa_dct,
                                 buffer=args.buf, gmxdata=args.ff, guess_elements=args.guess)
     except FileNotFoundError as e:
         print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
@@ -268,6 +280,121 @@ def get_nsa_mpt(args, only_nsa=False):
 def getmpt(args):
     get_nsa_mpt(args).write(args.mpt)
 
+def cpmdid(args):
+    
+    def print_idx(data, printer=print):
+        if args.print == 'table':
+            mimicpy.utils.strings.print_table(data, printer, new_line=(False if printer==print else True))
+            return
+        elif args.print == 'list':
+            col_len = 5
+            space_len = 6
+            indices = data['CPMD ID']
+            max_len = len(str(max(indices))) + 1
+            spaces = space_len if max_len <= space_len else max_len
+            txt = "No. of atoms: {}\n".format(len(indices))
+            for i, idx in enumerate(indices):
+                if i%col_len == 0:
+                    txt += '\n'
+                txt += "{:{}}".format(idx, spaces)
+        elif args.print == 'range':
+            indices = [int(i) for i in data['CPMD ID']]
+            indices.sort()
+            txt = ''
+            for i, j in itertools.groupby(enumerate(indices), lambda x: x[1] - x[0]): 
+                j = list(j)
+                if j[0][0] != 0: txt += '\n'
+                txt += "{} to {}".format(j[0][1], j[-1][1]) if j[0][1]!=j[-1][1] else str(j[0][1])
+        printer(txt)
+    
+    args.print = args.print.lower()
+    if args.print not in ['table', 'list', 'range']:
+        print('\nError: Invalid option for -print. Only table, list, range accepted! Exiting..\n')
+        sys.exit(1)
+    
+    try:
+        cpmd = mimicpy.CpmdScript.from_file(args.inp)
+        print("\n**Reading topology**\n")
+        top = mimicpy.Top(args.top, buffer=args.buf, gmxdata=args.ff)
+        id_conversion_dict = cpmd.gmx_to_cpmd_idx(top)
+        mpt = mimicpy.Mpt.from_top(top)
+    except mimicpy.utils.errors.ParserError as e:
+        print(e)
+        sys.exit(1)
+    except FileNotFoundError as e:
+        print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
+        sys.exit(1)
+        
+    readline.parse_and_bind('set editing-mode vi') # handle command history
+    
+    if args.sele:
+        print("Reading selection from {}".format(args.sele))
+
+        if not isfile(args.sele):
+            print("Error: Cannot find file {}! Exiting..\n".format(args.sele))
+            sys.exit(1)
+        else:
+            sele_txt =  mimicpy.utils.file_handler.read(args.sele)
+
+    else:
+        print("\nPlease enter selection below. Type q or quit to exit.")
+    
+        sele_txt = ''
+    
+        while not args.sele:
+            try:
+                user_input = input('> ')
+            except KeyboardInterrupt:
+                print("Exiting\n")
+                sys.exit()
+                
+            user_input = user_input.strip()
+            
+            if user_input in ['quit', 'q']: break
+            
+            try:
+                mpt.select(user_input)
+            except (mimicpy.utils.errors.SelectionError, mimicpy.utils.errors.MiMiCPyError) as error:
+                print(error)
+                continue
+                
+            sele_txt += user_input + '\n'
+    
+    data = {'Atom': [], 'Residue': [], 'CPMD ID': []}
+    
+    for line in sele_txt.splitlines():
+        try:
+            sele = mpt.select(line)
+        except (mimicpy.utils.errors.SelectionError, mimicpy.utils.errors.MiMiCPyError) as error:
+            print(error)
+            return None
+            
+        cpmd_idx = []
+        for i in sele.index:
+            cpmd_idx.append(str(id_conversion_dict[i]))
+        
+        mini_data = {'Atom': ["{} {}".format(a,b) for a,b in zip(sele.index, sele['name'])],\
+                'Residue': ["{} {}".format(a,b) for a,b in zip(sele['resid'], sele['resname'])], 'CPMD ID': cpmd_idx}
+        
+        data['Atom'] += mini_data['Atom']
+        data['Residue'] += mini_data['Residue']
+        data['CPMD ID'] += mini_data['CPMD ID']
+    
+    if not args.out:
+        print("\nThe CPMD IDs for the selected atoms are:\n")
+        print_idx(data)
+    elif args.out and data:
+        print("\n**Writing table of selected atoms**\n")
+        try:
+            with open(args.out, 'w') as f:
+                print_idx(data, f.write)
+        except FileNotFoundError as e:
+            print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
+            sys.exit(1)
+        except FileExistsError as e:
+            print(e)
+            sys.exit(1)
+
 def cpmd2coords(args):
     mpt = get_nsa_mpt(args)
     try:
@@ -290,12 +417,54 @@ def cpmd2coords(args):
         sys.exit(1)
 
     loader.close()
+    
+def geom2coords(args):
+    try:
+        cpmd = mimicpy.CpmdScript.from_file(args.inp)
+        print("\n**Reading topology**\n")
+        top = mimicpy.Top(args.top, buffer=args.buf, gmxdata=args.ff)
+        id_conversion_dict = cpmd.gmx_to_cpmd_idx(top)
+        mpt = mimicpy.Mpt.from_top(top)
+    except mimicpy.utils.errors.ParserError as e:
+        print(e)
+        sys.exit(1)
+    except FileNotFoundError as e:
+        print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
+        sys.exit(1)
+    
+    ext = 'cpmd geo'
+    if args.geom.split()[-1].lower() == 'xyz':
+        ext = 'xyz'
+    
+    print('')
+    loader = Loader('**Reading CPMD Geometry**')
+    
+    try:
+        geom = mimicpy.CoordsIO(args.geom, ext=ext, cpmd=cpmd, top=top)
+    except mimicpy.utils.errors.MiMiCPyError as e:
+        print(e)
+        loader.close(halt=True)
+        sys.exit(1)
+        
+    loader.close()
+    
+    print('')
+    loader = Loader('**Writing coordinates**')
+    
+    try:
+        mimicpy.CoordsIO(args.coords, mode='w').write(mpt.select('all'), geom.coords)
+    except mimicpy.utils.errors.MiMiCPyError as e:
+        print(e)
+        loader.close(halt=True)
+        sys.exit(1)
+
+    loader.close()
 
 def fixtop(args):
     nsa_dct = get_nsa_mpt(args, True)
     print("\n**Reading topology**\n")
     try:
-        top = mimicpy.Top(args.top, mode='w', buffer=args.buf, nonstandard_atomtypes=nsa_dct,
+        top = mimicpy.Top(args.top, buffer=args.buf, nonstandard_atomtypes=nsa_dct,
                         gmxdata=args.ff, guess_elements=args.guess)
     except FileNotFoundError as e:
         print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
@@ -306,7 +475,7 @@ def fixtop(args):
     print("\n**Writing fixed [ atomtypes ] section**\n")
 
     try:
-        top.write_atomtypes(args.fix, args.cls)
+        top.write_atomtypes(args.out, delete_atomtypes=args.cls)
     except FileNotFoundError as e:
         print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
         sys.exit(1)
@@ -315,19 +484,9 @@ def fixtop(args):
         sys.exit(1)
     return top
 
-def prepmm(args):
-    try:
-        mimicpy.Preparation.get_gmx_input(inp=args.mdp, qmatoms=args.qma, out=args.out)
-    except FileNotFoundError as e:
-        print('\n\nError: Cannot find file {}! Exiting..\n'.format(e.filename))
-        sys.exit(1)
-    except mimicpy.utils.errors.ScriptError as e:
-        print(e)
-        sys.exit(1)
-
 def main():
     print('\n \t                ***** MiMiCPy *****                  ')
-    print('\n \t For more information type mimicpy [subcommand] --help \n')
+    print('\n \t                Running version {}\n\tFor more information type mimicpy [subcommand] --help \n'.format(mimicpy.__version__))
 
     parser = argparse.ArgumentParser(prog='mimicpy')
     subparsers = parser.add_subparsers(title='valid subcommands',
@@ -340,7 +499,7 @@ def main():
     prepqm_input.add_argument('-top',
                               required=True,
                               help='topology file',
-                              metavar='[.top/.mpt]')
+                              metavar='[.top]')
     prepqm_input.add_argument('-coords',
                               required=True,
                               help='coordinate file',
@@ -358,10 +517,6 @@ def main():
                                default='index.ndx',
                                help='Gromacs index file',
                                metavar='[.ndx] (index.ndx)')
-    prepqm_output.add_argument('-fix',
-                               required=False,
-                               help='optional argument to specify fixed .itp file',
-                               metavar='[.itp]')
     prepqm_output.add_argument('-tpr',
                               required=False,
                               default='mimic.tpr',
@@ -382,11 +537,11 @@ def main():
                                required=False,
                                help='file containing pseudopotential information',
                                metavar='[.dat]')
-    prepqm_others.add_argument('-link',
+    prepqm_others.add_argument('-bound',
                               required=False,
                               default=False,
                               type=__str2bool,
-                              help='toggle guessing link atoms',
+                              help='toggle guessing boundary atoms',
                               metavar='(False)')
     prepqm_others.add_argument('-nsa',
                               required=False,
@@ -443,40 +598,16 @@ def main():
                               type=__str2int,
                               help='buffer size for reading input coordinates',
                               metavar='(1000)')
-    prepqm_others.add_argument('-cls',
-                              required=False,
-                              help='clear [ atomtypes ] sections from list of files if -fix option is enabled',
-                              nargs='*',
-                              metavar='.itp')
     parser_prepqm.set_defaults(func=prepqm)
     ##
     #####
-    parser_prepmm = subparsers.add_parser('prepmm',
-                                          help='create/fix GROMACS MDP script for MiMiC run')
-    prepmm_input = parser_prepmm.add_argument_group('options to specify input files')
-    prepmm_input.add_argument('-mdp',
-                              required=False,
-                              help='MDP template script',
-                              metavar='[.mdp]')
-    prepmm_input.add_argument('-qma',
-                              required=False,
-                              help='name of QM atoms group',
-                              metavar='QMatoms')
-    prepmm_output = parser_prepmm.add_argument_group('options to specify output')
-    prepmm_output.add_argument('-out',
-                               default='mimic.mdp',
-                               help='fixed MDP script for MiMiC run',
-                               metavar='[.mdp] (mimic.mdp)')
-    parser_prepmm.set_defaults(func=prepmm)
-    ##
-    #####
     parser_cpmd2coords = subparsers.add_parser('cpmd2coords',
                                           help='convert CPMD/MiMiC input to coordinates')
     cpmd2coords_input = parser_cpmd2coords.add_argument_group('options to specify input')
     cpmd2coords_input.add_argument('-top',
                               required=True,
                               help='topology file',
-                              metavar='[.top/.mpt]')
+                              metavar='[.top]')
     cpmd2coords_input.add_argument('-inp',
                               required=True,
                               help='CPMD input script with MIMIC/ATOMS sections',
@@ -517,7 +648,7 @@ def main():
                               help='GROMACS topology file',
                               metavar='[.top]')
     fixtop_output = parser_fixtop.add_argument_group('options to specify output files')
-    fixtop_output.add_argument('-fix',
+    fixtop_output.add_argument('-out',
                                default='atomtypes.itp',
                                help='fixed .itp file',
                                metavar='[.itp] (atomtypes.itp)')
@@ -537,9 +668,10 @@ def main():
                               metavar='')
     fixtop_others.add_argument('-cls',
                               required=False,
-                              help='clear atomtypes sections from list of files',
-                              nargs='*',
-                              metavar='.itp')
+                              default=False,
+                              type=__str2bool,
+                              help='toggle clear [ atomtypes ] sections from files',
+                              metavar='(False)')
     fixtop_others.add_argument('-buf',
                               required=False,
                               default=1000,
@@ -549,40 +681,77 @@ def main():
     parser_fixtop.set_defaults(func=fixtop)
     ##
     #####
-    parser_getmpt = subparsers.add_parser('getmpt',
-                                          help='create MiMiCPy topology from GROMACS topology')
-    getmpt_input = parser_getmpt.add_argument_group('options to specify input')
-    getmpt_input.add_argument('-top',
+    parser_cpmdid = subparsers.add_parser('cpmdid',
+                                          help='find the CPMD ID of a selection of atoms')
+    cpmdid_input = parser_cpmdid.add_argument_group('options to specify input files')
+    cpmdid_input.add_argument('-top',
                               required=True,
                               help='GROMACS topology file',
                               metavar='[.top]')
-    getmpt_output = parser_getmpt.add_argument_group('options to specify output')
-    getmpt_output.add_argument('-mpt',
-                               default='topol.mpt',
-                               help='MiMiCPy topology file',
-                               metavar='[.mpt] (topol.mpt)')
-    getmpt_others = parser_getmpt.add_argument_group('other options')
-    getmpt_others.add_argument('-guess',
+    cpmdid_input.add_argument('-inp',
                               required=False,
-                              default=True,
-                              type=__str2bool,
-                              help='toggle guessing atomic elements',
-                              metavar='(True)')
-    getmpt_others.add_argument('-nsa',
+                              help='CPMD input script',
+                              metavar='[.inp]')
+    cpmdid_others = parser_cpmdid.add_argument_group('options to specify output files')
+    cpmdid_others.add_argument('-sele',
                               required=False,
-                              help='list of non-standard atomtypes',
+                              help='file containing selection',
                               metavar='[.txt/.dat]')
-    getmpt_others.add_argument('-ff',
+    cpmdid_others.add_argument('-print',
+                              required=False,
+                              default='table',
+                              help='print CPMD IDs as table, list or range',
+                              metavar='table/list/range')
+    cpmdid_others.add_argument('-ff',
                               required=False,
                               help='path to force field data directory',
                               metavar='')
-    getmpt_others.add_argument('-buf',
+    cpmdid_others.add_argument('-buf',
                               required=False,
                               default=1000,
                               type=__str2int,
                               help='buffer size for reading input topology',
                               metavar='(1000)')
-    parser_getmpt.set_defaults(func=getmpt)
+    cpmdid_output = parser_cpmdid.add_argument_group('options to specify output')
+    cpmdid_output.add_argument('-out',
+                               required=False,
+                               help='file to write to output containing CPMD IDs',
+                               metavar='[.txt]')
+    parser_cpmdid.set_defaults(func=cpmdid)
+    ##
+    #####
+    parser_geom2coords = subparsers.add_parser('geom2coords',
+                                          help='convert a CPMD GEOMETRY (or a GEOMETRY.xyz) file to coordinates')
+    geom2coords_input = parser_geom2coords.add_argument_group('options to specify input files')
+    geom2coords_input.add_argument('-geom',
+                              required=False,
+                              help='CPMD GEOMETRY',
+                              metavar='[GEOMETRY/GEOMETRY.xyz]')
+    geom2coords_input.add_argument('-top',
+                              required=True,
+                              help='GROMACS topology file',
+                              metavar='[.top]')
+    geom2coords_input.add_argument('-inp',
+                              required=False,
+                              help='CPMD input script',
+                              metavar='[.inp]')
+    geom2coords_others = parser_geom2coords.add_argument_group('options to specify output files')
+    geom2coords_others.add_argument('-ff',
+                              required=False,
+                              help='path to force field data directory',
+                              metavar='')
+    geom2coords_others.add_argument('-buf',
+                              required=False,
+                              default=1000,
+                              type=__str2int,
+                              help='buffer size for reading input topology',
+                              metavar='(1000)')
+    geom2coords_output = parser_geom2coords.add_argument_group('options to specify output')
+    geom2coords_output.add_argument('-coords',
+                               default='GEOMETRY.gro',
+                               help='coordinate file from CPMD GEOMETRY',
+                               metavar='[.gro/.pdb] (GEOMETRY.gro)')
+    parser_geom2coords.set_defaults(func=geom2coords)
     ##
     args = parser.parse_args()
     if vars(args) == {}:
